indexing

	description: "Base class for various emc programs."


class

	EMC_BASE


inherit

	ANY

	EPX_CURRENT_PROCESS
		export
			{NONE} all
		end

	EPX_FILE_SYSTEM
		export
			{NONE} all
		end


feature -- Access

	configuration: CONFIGURATION
			-- Used configuration

	config_file_name: STRING
			-- Location of configuration file

	version: STRING is "0.7.4"

	copyright: STRING is "(c) 2003-2006 by Berend de Boer"


feature -- Configuration

	determine_config_location is
			-- Attempt to locate the "emc.config" file. If not found,
			-- terminate the program.
		do
			-- First location is current directory
			config_file_name := current_dir_config_file_name
			if not is_readable (config_file_name) then
				config_file_name := home_dir_config_file_name
				if not is_readable (config_file_name) then
					config_file_name := etc_dir_config_file_name
					if not is_readable (config_file_name) then
						stderr.write_string ("Cannot find location of emc.config.%N")
						stderr.write_string ("Location should be the current directory, ~/.emc/, or /usr/local/etc/emc/.%N")
						exit_with_failure
					end
				end
			end
		ensure
			config_file_name_set: config_file_name /= Void and then not config_file_name.is_empty
			found: is_readable (config_file_name)
		end

	read_configuration is
			-- Read emc configuration file.
		local
			file: STDC_TEXT_FILE
		do
			determine_config_location
			stdout.write_string ("Reading configuration from ")
			stdout.write_string (config_file_name)
			stdout.write_string ("...%N")
			create file.open_read (config_file_name)
			create configuration.make (file)
			file.close
			if not configuration.is_correct then
				stderr.write_string ("The configuration file contained an error: ")
				stderr.write_string (configuration.last_error)
				stderr.write_string ("%N")
				exit_with_failure
			end
		ensure
			configuration_not_void: configuration /= Void
		end

	check_configuration is
			-- Check if configuration information looks correct.
		require
			configuration_not_void: configuration /= Void
		do
			if configuration.password = Void then
				stdout.write_string ("Please enter your IMAP4 password for ")
				stdout.write_string (configuration.host)
				stdout.write_string (": ")
				stdin.read_line
				if stdin.last_string.is_empty then
					exit_with_failure
				else
					configuration.set_password (stdin.last_string)
				end
			end
		end


feature {NONE} -- Configuration

	current_dir_config_file_name: STRING is
			-- emc.config in current directory.
		once
			Result := current_directory + "/emc.config"
		ensure
			file_name_not_empty: Result /= Void and then not Result.is_empty
		end

	home_dir_config_file_name: STRING is
			-- emc.config in home directory.
		once
			create {STDC_PATH} Result.make_expand ("~/.emc/emc.config")
		ensure
			file_name_not_empty: Result /= Void and then not Result.is_empty
		end

	etc_dir_config_file_name: STRING is
			-- emc.config in /usr/local/etc/emc/ directory.
		once
			Result := "/usr/local/etc/emc/emc.config"
		ensure
			file_name_not_empty: Result /= Void and then not Result.is_empty
		end


feature -- Connection

	connect_to_mail_server: EPX_IMAP4_CLIENT is
			-- Connect to server.
		do
			create Result.make_with_port (configuration.host, configuration.port, configuration.secure)
			if not Result.is_open then
				stderr.put_string ("Could not make a connection to server ")
				stderr.put_string (configuration.host)
				stderr.put_character (':')
				stderr.put_string (Result.port.out)
				stderr.put_string (".%N")
				if Result.error_message /= Void then
					stderr.put_string ("Reason: ")
					stderr.put_string (Result.error_message)
					stderr.put_character ('%N')
				else
					stderr.put_string ("No additional reason given.%N")
				end
				exit_with_failure
			end
			Result.login (configuration.login_name, configuration.password)
			if not Result.response.is_ok then
				stderr.put_string ("Login to server: ")
				stderr.put_string (configuration.host)
				stderr.put_character (':')
				stderr.put_string (Result.port.out)
				stderr.put_string (" as user ")
				stderr.put_string (configuration.login_name)
				stderr.put_string (" failed.%N")
				exit_with_failure
			end
		ensure
			client_not_void: Result /= Void
			client_authenticated: Result.state.is_authenticated
		end


feature -- Checks

	get_writable_mailbox (a_client: EPX_IMAP4_CLIENT; a_mailbox_name: STRING): EPX_IMAP4_MAILBOX is
			-- Check if `a_mailbox_name' is a selectable mailbox on
			-- `a_client'. Return that mailbox.
			-- Halt program, if not.
		require
			a_client_not_void: a_client /= Void
			a_client_authenticated: a_client.state.is_authenticated
			valid_mailbox_name: a_client.is_valid_mailbox_name (a_mailbox_name)
		do
			a_client.select_mailbox (a_mailbox_name)
			if a_client.response.is_ok then
				Result := a_client.response.current_mailbox
				a_client.close_mailbox
			else
				a_client.examine (a_mailbox_name)
				if a_client.response.is_ok then
					stderr.put_string ("Fatal error: write access to mailbox ")
					stderr.put_string (a_mailbox_name)
					stderr.put_string (" denied.%N")
					a_client.close_mailbox
				else
					stderr.put_string ("Fatal error: cannot access mailbox: ")
					stderr.put_string (a_mailbox_name)
					stderr.put_string (". It does not exist or access is denied.%N")
				end
				exit_with_failure
			end
		ensure
			state_remains_authenticated: a_client.state.is_authenticated
		end

end
