indexing

	description: "Class that scans email and builds a hash table for the tokens."


class

	CLASSIFY


inherit

	EMC_BASE

	KL_SHARED_ARGUMENTS
		export
			{NONE} all
		end

	KL_DOUBLE_ROUTINES
		export
			{NONE} all
		end

	TOKENIZE_ROUTINES

	ST_FORMATTING_ROUTINES
		export
			{NONE} all
		end


creation

	make


feature {NONE} -- Initialization

	make is
		do
			stdout.put_string ("mail classifier " + version + "  " + copyright + ".%N")
			parse_arguments
			read_configuration
			check_configuration
			if mode = Dry_run then
				stdout.put_string ("(dry-run mode)%N")
			end
			stdout.put_string ("Reading probabilities...%N")
			determine_probabilities_location
			create token_probabilities.make (probabilities_file_name)

			-- Check to see if `new_word_spaminess' if set, if not, get
			-- it from probabilities.txt.
			if configuration.new_word_spaminess = 0 then
				configuration.set_new_word_spaminess (token_probabilities.no_data_points_value)
			end

			if verbosity >= 1 then
				stdout.put_string (format("Messages with a spam probability equal to or larger than $.2f are considered to be spam%N", <<double_cell (configuration.threshold)>>))
				if
					configuration.lower_minimum_deviation /= 0 or else
					configuration.upper_minimum_deviation /= 0
				then
					stdout.put_string (format("Words need to have a spaminess probability less than $.2f and larger than $.2f to be counted%N", <<double_cell (configuration.lower_minimum_deviation), double_cell (configuration.upper_minimum_deviation)>>))
				end
			end
			if verbosity >= 2 then
				stdout.put_string (format("Number of tokenized spam messages   : $i%N", <<integer_cell (token_probabilities.spam_messages_count)>>))
				stdout.put_string (format("Number of tokenized nonspam messages: $i%N", <<integer_cell (token_probabilities.nonspam_messages_count)>>))
				stdout.put_string (format("Probability of new word to be spam  : $f%N", <<double_cell (configuration.new_word_spaminess)>>))
				stdout.put_string (format("Strength factor (s constant)        : $f%N", <<double_cell (configuration.strength)>>))
			end

			stdout.put_string ("Starting classification...%N")
			classify_mailbox (configuration.scan_mailbox, configuration.spam_mailbox, configuration.nonspam_mailbox)
		end


feature -- Status

	is_read_only: BOOLEAN is
			-- Only read-only access to mail server.
		do
			Result := mode /= Classify
		ensure
			classify_mode_implies_write: mode = Classify implies not Result
		end


feature -- Access

	Algorithms,
	Classify,
	Dry_run: INTEGER is unique
			-- Valid `mode' states.
			-- @@BdB: ugly, need to become separate class.

	mode: INTEGER
			-- List of modes.

	verbosity: INTEGER
			-- How much logging is done.


feature -- Argument parsing

	parse_arguments is
		local
			arg: STRING
			i: INTEGER
		do
			mode := Classify
			from
				i := 1
			until
				i > Arguments.argument_count
			loop
				arg := Arguments.argument (i)
				if arg.is_equal ("--dry-run") then
					mode := Dry_run
				elseif arg.is_equal ("--algorithms") then
					mode := Algorithms
				elseif arg.is_equal ("-v") then
					verbosity := verbosity + 1
				elseif arg.is_equal ("-vv") then
					verbosity := verbosity + 2
				elseif arg.is_equal ("-vvv") then
					verbosity := verbosity + 3
				elseif arg.is_equal ("-vvvv") then
					verbosity := verbosity + 4
				elseif arg.is_equal ("-h") or else arg.is_equal ("--help") then
					print_help
					exit_with_success
				else
					stderr.put_string ("Unrecognized argument '")
					stderr.put_string (arg)
					stderr.put_string (".%N")
					exit_with_failure
				end
				i := i + 1
			end
		end

	print_help is
		do
			stderr.put_string ("Recognized arguments:%N%N")
			stderr.put_string ("--dry-run  Show only what will be done, don't actually do it.%N")
			stderr.put_string ("-h         Display this help.%N")
			stderr.put_string ("-v         Increase verbosity.%N%N")
		end


feature -- Configuration

	determine_probabilities_location	is
			-- Attempt to locate the "probabilities.txt" file. If not found,
			-- terminate the program.
		do
			create probabilities_file_name.make_from_string (config_file_name)
			probabilities_file_name.parse (Void)
			probabilities_file_name.set_basename ("probabilities.txt")
			if not is_readable (probabilities_file_name) then
				stderr.put_string ("The file probabilities.txt could not be read.%N")
				stderr.put_string ("This file is generated by the scan utility. Make sure you have run that first.%N")
				stderr.put_string ("Put the generated probabilities.txt in the same directory as your emc.config%N")
				stderr.put_string ("file. That directory is ")
				stderr.put_string (probabilities_file_name.directory)
				stderr.put_string ("/%N")
				exit_with_failure
			end
		ensure
			probabilities_file_name_set: probabilities_file_name /= Void and then not probabilities_file_name.is_empty
			found: is_readable (probabilities_file_name)
		end

	probabilities_file_name: STDC_PATH
			-- Path to "probabilities.txt"


feature -- Classification

	classify_mailbox (a_mailbox: MAILBOX; a_spam_mailbox_name, a_nonspam_mailbox_name: STRING) is
			-- Reads through all unseen messages in `a_mailbox_name' and
			-- moves them to `a_spam_mailbox_name' or
			-- `a_nonspam_mailbox_name'.  If `a_spam_mailbox_name' or
			-- `a_nonspam_mailbox_name' is Void, the message is not
			-- moved.
		require
			a_mailbox_not_void: a_mailbox /= Void
		local
			client: EPX_IMAP4_CLIENT
			message_number: INTEGER
			percentage_done,
			prev_percentage_done: INTEGER
			total_messages_to_scan: INTEGER
			parser: EPX_MIME_PARSER
			spam_mailbox,
			nonspam_mailbox: EPX_IMAP4_MAILBOX
			spam_percentage,
			nonspam_percentage: DOUBLE
		do
			stdout.put_string ("Classifying ")
			stdout.put_string (a_mailbox.name)
			stdout.put_string ("%N")
			stdout.put_string ("Connecting to server ")
			stdout.put_string (configuration.host)
			if configuration.port /= 0 then
				stdout.put_character (':')
				stdout.put_string (configuration.port.out)
			end
			stdout.put_string (" as user ")
			stdout.put_string (configuration.login_name)
			stdout.put_string ("...%N")
			client := connect_to_mail_server

			if not is_read_only then
				if a_spam_mailbox_name /= Void then
					spam_mailbox := get_writable_mailbox (client, a_spam_mailbox_name)
				end
				if a_nonspam_mailbox_name /= Void then
					nonspam_mailbox := get_writable_mailbox (client, a_nonspam_mailbox_name)
				end
			end

			if is_read_only then
				client.examine (a_mailbox.name)
			else
				client.select_mailbox (a_mailbox.name)
			end
			if not client.response.is_ok then
				stderr.put_string ("Fatal error: cannot read mailbox: ")
				stderr.put_string (a_mailbox.name)
				stderr.put_string (".%N")
				exit_with_failure
			end
			if is_read_only then
				assert ("Read-only mailbox.", not client.response.current_mailbox.is_writable)
			else
				assert ("Writable mailbox.", client.response.current_mailbox.is_writable)
			end
			stdout.put_string ("Number of messages: ")
			stdout.put_string (client.response.current_mailbox.count.out)
			stdout.put_string ("%N")

			create parser.make
			from
				prev_percentage_done := -1
				total_messages_to_scan := client.response.current_mailbox.count
				message_number := a_mailbox.lowest_sequence
				--i := client.response.current_mailbox.count - 10
			until
				message_number > total_messages_to_scan or else
				message_number > client.response.current_mailbox.count
			loop
				percentage_done := (message_number * 100) // total_messages_to_scan
				if percentage_done /= prev_percentage_done then
					fd_stdout.put_string ("Completed ")
					fd_stdout.put_string (percentage_done.out)
					fd_stdout.put_string ("%%%R")
					prev_percentage_done := percentage_done
				end
				classify_message (client, parser, message_number)
				message_number := message_number + 1
			end

			fd_stdout.put_string ("Completed 100%%%N")

			client.close_mailbox
			if spam_mailbox /= Void then
				mark_new_messages_unseen (client, spam_mailbox)
			end
			if nonspam_mailbox /= Void then
				mark_new_messages_unseen (client, nonspam_mailbox)
			end
			if verbosity >= 1 then
				if spam_messages + nonspam_messages > 0 then
					spam_percentage := (spam_messages / (spam_messages + nonspam_messages)) * 100
					nonspam_percentage := (nonspam_messages / (spam_messages + nonspam_messages)) * 100
				else
					spam_percentage := 0
					nonspam_percentage := 0
				end
				fd_stdout.put_string (format ("Messages classified as spam    : $3i ($3.1f%%)%N", <<integer_cell (spam_messages), double_cell (spam_percentage)>>))
				fd_stdout.put_string (format ("Messages classified as not spam: $3i ($3.1f%%)%N", <<integer_cell (nonspam_messages), double_cell (nonspam_percentage)>>))
			end
			client.logout
			client.close
		end


feature {NONE} -- Classification

	spam_messages: INTEGER
			-- Number of messages classified as spam

	nonspam_messages: INTEGER
			-- Number of messages classified as non-spam

	act_on_message (
				client: EPX_IMAP4_CLIENT;
				a_sequence_number: INTEGER;
				a_spam_probability: DOUBLE;
				a_message_seen: BOOLEAN) is
			-- Depending on `a_spam_probability', move message or remove
			-- the \Seen flag if it was previously unseen.
		require
			client_not_void: client /= Void
			mailbox_selected: client.state.is_selected
			sequence_number_valid: client.is_valid_sequence_number (a_sequence_number)
		do
			if a_spam_probability >= configuration.threshold then
				move_message (client, a_sequence_number, configuration.spam_mailbox, a_message_seen)
				spam_messages := spam_messages + 1
			else
				move_message (client, a_sequence_number, configuration.nonspam_mailbox, a_message_seen)
				nonspam_messages := nonspam_messages + 1
			end
		end

	classify_message (client: EPX_IMAP4_CLIENT; parser: EPX_MIME_PARSER; a_message_number: INTEGER) is
			-- Retrieve and classify the given message. Call
			-- `act_on_message' to take the appropriate action.
		require
			client_not_void: client /= Void
			parser_not_void: parser /= Void
			message_number_valid: client.is_valid_sequence_number (a_message_number)
		local
			seen: BOOLEAN
			probability: DOUBLE
			tokens: DS_HASH_TABLE [INTEGER, STRING]
		do
			client.fetch (a_message_number.out, "(FLAGS RFC822.SIZE)")
			if client.response.is_ok then
				seen := client.response.current_message.is_seen
				if not seen then
					if verbosity >= 2 then
						stdout.put_string ("Message number: ")
						stdout.put_string (a_message_number.out)
						stdout.put_string ("%N")
					end
					if client.response.current_message.size < configuration.scan_mailbox.max_message_size then
						client.fetch_header (a_message_number)
						assert ("Fetch ok", client.response.is_ok)
						client.fetch_body (a_message_number)
						assert ("Fetch body ok", client.response.is_ok)
						tokens := tokenize_message (a_message_number, parser, client.response.current_message.header, client.response.current_message.body)
						inspect mode
						when Algorithms then
							show_spam_probability (tokens, parser.part)
						when Classify, Dry_run then
							if is_read_only or else verbosity >= 1 then
								stdout.put_string ("%NSubject: ")
								if parser.part.header.has (field_name_subject) then
									stdout.put_string (parser.part.header.item (field_name_subject).value)
								end
								stdout.put_string ("%N")
							end
							probability := spam_probability (tokens)
							if is_read_only or else verbosity >= 1 then
								stdout.put_string ("Spam probability: ")
								stdout.put_string (probability.out)
								stdout.put_string (" => message is")
								if probability < configuration.threshold then
									stdout.put_string (" not")
								end
								stdout.put_string (" spam%N")
							end
							act_on_message (client, a_message_number, probability, seen)
						end
					else
						stderr.put_string ("Skipped large message: ")
						stderr.put_string (client.response.current_message.size.out)
						stderr.put_string (" bytes.%N")
					end
				end
			else
				stderr.put_string ("Failure to retrieve header: ")
				if client.response.response_text /= Void then
					stderr.put_string (client.response.response_text)
				end
				stderr.put_string ("%N")
			end
		end

	mark_new_messages_unseen (client: EPX_IMAP4_CLIENT; a_mailbox: EPX_IMAP4_MAILBOX) is
			-- Give all messages added after `a_mailbox'.`count' the
			-- \Unseen flag.
		require
			client_not_void: client /= Void
			authenticated:
				client.state.is_authenticated or else
				client.state.is_selected
			a_mailbox_not_void: a_mailbox /= Void
		local
			last_count,
			new_count: INTEGER
			i: INTEGER
		do
			last_count := a_mailbox.count
			client.select_mailbox (a_mailbox.name)
			new_count := a_mailbox.count
			from
				i := last_count + 1
			until
				i > new_count
			loop
				client.mark_unseen (i)
				i := i + 1
			end
		end

	move_message (
				client: EPX_IMAP4_CLIENT;
				a_sequence_number: INTEGER;
				to_mailbox_name: STRING;
				a_message_seen: BOOLEAN) is
			-- Move message to mailbox `to_mailbox_name' if this is not
			-- Void. If message is not moved, we try to remove the \Seen
			-- flag in case not `a_message_seen'.
		require
			client_not_void: client /= Void
			mailbox_selected: client.state.is_selected
			sequence_number_valid: client.is_valid_sequence_number (a_sequence_number)
			to_mailbox_name_valid:
				to_mailbox_name = Void or else
				client.is_valid_mailbox_name (to_mailbox_name)
		do
			if to_mailbox_name = Void then
				if not a_message_seen then
					if not is_read_only then
						client.mark_unseen (a_sequence_number)
					end
				end
			else
				if is_read_only then
					stdout.put_string ("Moving message ")
					stdout.put_string (a_sequence_number.out)
					stdout.put_string (" to ")
					stdout.put_string (to_mailbox_name)
					stdout.put_string ("%N")
				else
					client.copy_message (a_sequence_number, to_mailbox_name)
					if not client.response.is_ok then
						stderr.put_string (client.response.response_text)
						stderr.put_string ("%N")
					end
					assert ("Copy successful.", client.response.is_ok)
					client.delete_message (a_sequence_number)
					if not client.response.is_ok then
						stderr.put_string (client.response.response_text)
						stderr.put_string ("%N")
					end
					assert ("Delete successful.", client.response.is_ok)
				end
			end
		end

	tokenize_message (a_message_number: INTEGER; parser: EPX_MIME_PARSER; a_message_header, a_message_body: STRING): DS_HASH_TABLE [INTEGER, STRING] is
			-- The tokens when tokenizing the given message using
			-- `parser'.
		require
			parser_not_void: parser /= Void
			message_header_not_void: a_message_header /= Void and then not a_message_header.is_empty
			message_body_not_void: a_message_body /= Void
		local
			body_tokenizer: TOKENIZER
			header_parsed: BOOLEAN
		do
			create body_tokenizer.make
			parser.set_input_buffer (parser.new_string_buffer (a_message_header))
			parser.parse_header
			header_parsed :=
				not parser.syntax_error and then
				parser.parsing_errors = 0
			debug ("emc")
				if not header_parsed then
					stderr.put_string ("%NError parsing message header: ")
					stderr.put_string (a_message_number.out)
					stderr.put_string ("%N")
					stderr.put_string ("==========begin==========%N")
					stderr.put_string (a_message_header)
					stderr.put_string ("==========end==========%N%N")
				end
			end
			parser.set_input_buffer (parser.new_string_buffer (a_message_body))
			parser.reset_parsing_errors
			parser.parse_body
			debug ("emc")
				if
					parser.syntax_error or else
					parser.parsing_errors > 0
				then
					stderr.put_string ("%NError parsing message body: ")
					stderr.put_string (a_message_number.out)
					stderr.put_string ("%N")
					stderr.put_string ("==========begin==========%N")
					if a_message_body.count <= 1024 then
						stderr.put_string (a_message_body)
					else
						stderr.put_string (a_message_body.substring (1, 1024))
					end
					stderr.put_string ("==========end==========%N%N")
				end
			end
			tokenize_text_bodies (body_tokenizer, a_message_number, parser.part)
			tokenize_header (body_tokenizer, parser.part.header)
			if not header_parsed then
				tokenize_bad_header (body_tokenizer)
			end
			Result := body_tokenizer.hash_table
		end


feature -- Spam

	show_spam_probability (a_tokens: DS_HASH_TABLE [INTEGER, STRING]; part: EPX_MIME_PART) is
			-- Show probability if message `part' being spam.
		require
			tokens_not_void: a_tokens /= Void
			part_not_void: part /= Void
		local
			probability1,
			probability2,
			probability3: DOUBLE
			pw: DS_ARRAYED_LIST [TOKEN_PROBABILITY]
		do
			-- Calculate the probability for every token.
			pw := assign_probabilities (token_probabilities, a_tokens, configuration.new_word_spaminess)

			-- Limit the number of tokens to consider
			-- In my experience, it detoriates the results
			--pw := limit_list_to_number (pw, 100)

			-- Remove probabilities not in range
			if
				configuration.lower_minimum_deviation /= 0 or else
				configuration.upper_minimum_deviation /= 0
			then
				pw := limit_list_to_probabilities (pw, configuration.lower_minimum_deviation, configuration.upper_minimum_deviation)
			end

			-- Use three different formula's
			probability1 := spam_probability_gr1 (pw)
			probability2 := spam_probability_gr2 (configuration.strength, configuration.new_word_spaminess, pw)
			probability3 := spam_probability_gr3 (configuration.strength, configuration.new_word_spaminess, token_probabilities.spam_messages_count, token_probabilities.nonspam_messages_count, pw)
			print ("Subject: ")
			if part.header.has ("Subject") then
				print (part.header.item ("Subject").value)
			else
				print ("<no subject>%N")
			end
			print ("%N")
			print ("  Paul Graham + Gary Robinson   : ")
			print (probability1)
			print ("%N")
			print ("  Gary Robinson improved formula: ")
			print (probability2)
			print ("%N")
			print ("  Greg Louis                    : ")
			print (probability3)
			print ("%N")
		end

	spam_probability (a_tokens: DS_HASH_TABLE [INTEGER, STRING]): DOUBLE is
			-- The probability of tokens in `body_tokenizer' being
			-- spam as a value between -1 (no spam) and 1 (spam).
		require
			tokens_not_void: a_tokens /= Void
		local
			pw: DS_ARRAYED_LIST [TOKEN_PROBABILITY]
		do
			-- Calculate the probability for every token.
			pw := assign_probabilities (token_probabilities, a_tokens, configuration.new_word_spaminess)

			-- Remove probabilities not in range
			if
				configuration.lower_minimum_deviation /= 0 or else
				configuration.upper_minimum_deviation /= 0
			then
				pw := limit_list_to_probabilities (pw, configuration.lower_minimum_deviation, configuration.upper_minimum_deviation)
			end

			-- Optionally show list of tokens and their probabilities
			if verbosity >= 3 then
				show_token_probabilities (pw)
			end

			-- Calculate the spaminess
			Result := spam_probability_gr3 (configuration.strength, configuration.new_word_spaminess, token_probabilities.spam_messages_count, token_probabilities.nonspam_messages_count, pw)
		end


feature {NONE} -- Probability calculation

	assign_probabilities (probabilities: PROBABILITY_TABLE; tokens: DS_HASH_TABLE [INTEGER, STRING]; x: DOUBLE): DS_ARRAYED_LIST [TOKEN_PROBABILITY] is
			-- Assign a probability to every token. This gives a list of
			-- p(w)'s according to Gary Robinson's notation.
			-- `x' is the value to assign when the token in `tokens' does
			-- not occur in `probabilities'.
		require
			probabilities_not_void: probabilities /= Void
			tokens_not_void: tokens /= Void
			x_is_probability: x >= 0 and then x <= 1
		local
			token_probability: TOKEN_PROBABILITY
		do
			create Result.make (tokens.count)
			from
				tokens.start
			until
				tokens.after
			loop
				probabilities.tokens.search (tokens.key_for_iteration)
				if probabilities.tokens.found then
					token_probability := probabilities.tokens.found_item
				else
					create token_probability.make (tokens.key_for_iteration, x, 0, 0)
				end
				Result.put_last (token_probability)
				tokens.forth
			end
		ensure
			list_not_void: Result /= Void
			all_tokens_assigned: Result.count = tokens.count
		end

	limit_list_to_number (a_list: DS_ARRAYED_LIST [TOKEN_PROBABILITY]; a_maximum: INTEGER): DS_ARRAYED_LIST [TOKEN_PROBABILITY] is
			-- Get `a_maximum' tokens that have the highest probability
			-- in `probabilities' (highest probability of being spam).
		require
			a_list_not_void: a_list /= Void
			positive_maximum: a_maximum >= 1
		local
			comparator: KL_COMPARABLE_COMPARATOR [TOKEN_PROBABILITY]
			sorter: DS_QUICK_SORTER [TOKEN_PROBABILITY]
			i: INTEGER
		do
			create comparator.make
			create sorter.make (comparator)
			sorter.sort (a_list)
			create Result.make (a_maximum)
			from
				a_list.finish
				i := 1
			until
				i > a_maximum or else
				a_list.before
			loop
				Result.put_last (a_list.item_for_iteration)
				a_list.back
				i := i + 1
			end
		ensure
			items_in_new_list: Result.count = a_maximum.min (a_list.count)
			no_more_items_than_maximum: Result.count <= a_maximum
		end

	limit_list_to_probabilities (a_list: DS_ARRAYED_LIST [TOKEN_PROBABILITY]; a_lower_range, an_upper_range: DOUBLE): DS_ARRAYED_LIST [TOKEN_PROBABILITY] is
			-- Return a list of tokens that have a probability that is
			-- not in the `a_lower_range'..`an_upper_range' range.
		require
			a_list_not_void: a_list /= Void
			lower_range_valid: a_lower_range >= 0 and then a_lower_range <= 1
			upper_range_valid: an_upper_range >= 0 and then an_upper_range <= 1
			valid_range: a_lower_range <= an_upper_range
		local
			data_point: TOKEN_PROBABILITY
		do
			if verbosity >= 4 then
				stdout.put_string ("Skipped tokens that are in the minimum deviation range:%N")
			end
			create Result.make (a_list.count)
			from
				a_list.start
			until
				a_list.after
			loop
				data_point := a_list.item_for_iteration
				if
					data_point.probability <= a_lower_range or else
					data_point.probability >= an_upper_range
				then
					Result.put_last (data_point)
				elseif verbosity >= 4 then
					stdout.put_string (format ("  $32s  $.6f%N", <<a_list.item_for_iteration.token, double_cell (a_list.item_for_iteration.probability)>>))
				end
				a_list.forth
			end
		ensure
			no_more_items_than_in_original: Result.count <= a_list.count
			probabilities_outside_range: True
		end

	spam_probability_gr1 (pw: DS_ARRAYED_LIST [TOKEN_PROBABILITY]): DOUBLE is
			-- Calculate spaminess given a set of probabilities in `pw'.
			-- `pw' should contain every token in the message.
			-- We follow the first formula given by Gary Robinson, where
			-- he improves Paul Graham's original formula.
		require
			pw_not_void: pw /= Void
		local
			P, Q, S: DOUBLE
			n: INTEGER
		do
			-- calculate P, the spaminess
			-- Formula: P = 1 - ((1-p1)*(1-p2)*...*(1-pn))^(1/ n)
			from
				P := 1
				pw.start
			until
				pw.after
			loop
				P := P * (1 - pw.item_for_iteration.probability)
				pw.forth
			end
			--P := 1 - P ^ (1 / pw.count)
			n := pw.count
			if n = 0 then
				P := 0.5
			else
				P := 1 - nth_root (P, n)
			end

			-- calculate Q, the non-spamminess
			-- Formula: Q = 1 - (p1*p2*...*pn)^(1/ n)
			from
				Q := 1
				pw.start
			until
				pw.after
			loop
				Q := Q * pw.item_for_iteration.probability
				pw.forth
			end
			--Q := 1 - Q ^ (1 / pw.count)
			n := pw.count
			if n = 0 then
				Q := 0.5
			else
				Q := 1 - nth_root (Q, pw.count)
			end

			-- calculate S
			S := (P - Q) / (P + Q)

			Result := S
		end


	spam_probability_gr2 (strength, x: DOUBLE; pw: DS_ARRAYED_LIST [TOKEN_PROBABILITY]): DOUBLE is
			-- Calculate spaminess given a set of probabilities (data
			-- points) in `pw'. We follow the improved formula given by
			-- Gary Robinson.
			-- `strenth' is the strength we want to give to our
			-- background information.
			-- `x' is our assumed probability, based on our general
			-- backround information, that a word we don't have any other
			-- experience of will first appear in a spam.
		require
			pw_not_void: pw /= Void
		local
			fw: ARRAY [DOUBLE]
			P, Q, S: DOUBLE
			data_point: TOKEN_PROBABILITY
			n: INTEGER
			i: INTEGER
			sum: DOUBLE
		do
			-- Calculate f(w) for every word
			create fw.make (1, pw.count)
			from
				pw.start
				i := 1
			until
				pw.after
			loop
				data_point := pw.item_for_iteration
				-- `n' is number of data points for a given token
				n := data_point.bad_count + data_point.good_count
				fw.put (((strength * x) + (n * data_point.probability)) / (strength + n), i)
				i := i + 1
				pw.forth
			end

			-- Now combine the probabilities according to the
			-- geometric-mean calculation:
			--   P = 1 - exp(sum(ln(1-f(w))) / n)
			--   Q = 1 - exp(sum(ln(f(w))) / n)
			--   S = (1 + (P-Q) / (P+Q)) / 2

			-- calculate P, the spaminess
			from
				sum := 0
				i := fw.lower
			until
				i > fw.upper
			loop
				sum := sum + log (1-fw.item (i))
				i := i + 1
			end
			n := pw.count
			if n = 0 then
				P := 0.5
			else
				P := 1 - exp (sum / n)
			end

			-- calculate Q, the non-spamminess
			from
				sum := 0
				i := fw.lower
			until
				i > fw.upper
			loop
				sum := sum + log (fw.item (i))
				i := i + 1
			end
			n := pw.count
			if n = 0 then
				Q := 0.5
			else
				Q := 1 - exp (sum / n)
			end

			-- calculate S
			S := (P - Q) / (P + Q)

			Result := S
		end

	spam_probability_gr3 (strength, x: DOUBLE; spam_messages_count, nonspam_messages_count: INTEGER; pw: DS_ARRAYED_LIST [TOKEN_PROBABILITY]): DOUBLE is
			-- Calculate spaminess given a set of probabilities in `pw'.
			-- We follow the f(w) formula given by Greg Louis.
		require
			spam_messages_count_positive: spam_messages_count > 0
			nonspam_messages_count_positive: nonspam_messages_count > 0
			pw_not_void: pw /= Void
		local
			fw: ARRAY [DOUBLE]
			P, Q, S: DOUBLE
			data_point: TOKEN_PROBABILITY
			n: INTEGER
			i: INTEGER
			sum: DOUBLE
			scale_factor: DOUBLE
		do
			scale_factor := spam_messages_count / nonspam_messages_count

			-- Calculate f(w) for every word. Formula:
			--   f(w) = (s * x + badcount) / (s + badcount + goodcount * scale_factor)
			create fw.make (1, pw.count)
			from
				pw.start
				i := 1
			until
				pw.after
			loop
				data_point := pw.item_for_iteration
				fw.put (((strength * x) + data_point.bad_count) / (strength + data_point.bad_count + (data_point.good_count * scale_factor)), i)
				i := i + 1
				pw.forth
			end

			-- Now combine the probabilities according to the
			-- geometric-mean calculation:
			--   P = 1 - exp(sum(ln(1-f(w))) / n)
			--   Q = 1 - exp(sum(ln(f(w))) / n)
			--   S = (1 + (P-Q) / (P+Q)) / 2

			-- calculate P, the spaminess
			from
				sum := 0
				i := fw.lower
			until
				i > fw.upper
			loop
				sum := sum + log (1-fw.item (i))
				i := i + 1
			end
			n := pw.count
			if n = 0 then
				P := 0.5
			else
				P := 1 - exp (sum / n)
			end

			-- calculate Q, the non-spamminess
			from
				sum := 0
				i := fw.lower
			until
				i > fw.upper
			loop
				sum := sum + log (fw.item (i))
				i := i + 1
			end
			n := pw.count
			if n = 0 then
				Q := 0.5
			else
				Q := 1 - exp (sum / n)
			end

			-- calculate S
			S := (P - Q) / (P + Q)

			Result := S
		end


feature -- Verbosity

	show_token_probabilities (a_list: DS_ARRAYED_LIST [TOKEN_PROBABILITY]) is
			-- Write probabilities of all words in `a_list' to `stderr'.
		local
			data_point: TOKEN_PROBABILITY
		do
			stdout.put_string ("Tokens in this message:%N")
			from
				a_list.start
			until
				a_list.after
			loop
				data_point := a_list.item_for_iteration
				stdout.put_string (format ("  $32s  $.6f%N", <<data_point.token, double_cell (data_point.probability)>>))
				a_list.forth
			end
		end


feature -- Error handling

	assert (msg: STRING; ok: BOOLEAN) is
			-- Stop if not `ok'.
		do
			if not ok then
				exceptions.raise (msg)
			end
		end


feature {NONE} -- Implementation

	token_probabilities: PROBABILITY_TABLE


invariant

	token_probabilities_not_void: token_probabilities /= Void
	configuration_not_void: configuration /= Void

end
