indexing

	description: "Class that reads a probability file."


class

	PROBABILITY_TABLE


inherit

	ANY

	EPX_CURRENT_PROCESS
		export
			{NONE} all
		end

	EPX_FILE_SYSTEM
		export
			{NONE} all;
			{ANY} is_readable
		end

	EPX_STRING_HELPER
		export
			{NONE} all
		end


creation

	make


feature {NONE} -- Initialization

	make (a_file_name: STRING) is
			-- Initialize probability table from existing file.
		require
			file_readable: is_readable (a_file_name)
		do
			load (a_file_name)
		end


feature -- Load/save

	load (a_file_name: STRING) is
			-- Load probability table from existing file.
		require
			file_readable: is_readable (a_file_name)
		local
			file: STDC_TEXT_FILE
			ar: ARRAY [STRING]
			probability: DOUBLE
			token: STRING
			bad_count,
			good_count: INTEGER
			token_probability: TOKEN_PROBABILITY
			line: INTEGER
		do
			create file.open_read (a_file_name)

			-- First line is spam msgs and nonspam msgs
			line := 1
			file.read_line
			ar := split_on (file.last_string, ' ')
			assert (line, "Spam and nonspam message count", ar.count = 2)
			assert (line, "First word is an integer count", ar.item (ar.lower).is_integer)
			assert (line, "Second word is an integer count", ar.item (ar.upper).is_integer)
			spam_messages_count := ar.item (ar.lower).to_integer
			nonspam_messages_count := ar.item (ar.upper).to_integer

			-- Second line is x
			line := line + 1
			file.read_line
			assert (line, "Line is a valid DOUBLE", file.last_string.is_double)
			no_data_points_value := file.last_string.to_double

			-- Third line is number of tokens
			line := line + 1
			file.read_line
			assert (line, "Line is a valid count", file.last_string.is_integer)
			create tokens.make (file.last_string.to_integer)

			-- Rest of lines are tokens
			from
				line := line + 1
				file.read_line
			until
				file.eof
			loop
				file.last_string.left_adjust
				file.last_string.right_adjust
				ar := split_on (file.last_string, ' ')
				assert (line, "Correct token probability line", ar.count = 4)
				assert (line, "First word is a valid DOUBLE", ar.item (ar.lower).is_double)
				probability := ar.item (ar.lower).to_double
				token := ar.item (ar.lower + 1)
				assert (line, "Second word is a non-empty token", not token.is_empty)
				assert (line, "Third word is an integer count", ar.item (ar.lower + 2).is_integer)
				assert (line, "Third word is an integer count", ar.item (ar.lower + 3).is_integer)
				bad_count := ar.item (ar.lower + 2).to_integer
				good_count := ar.item (ar.lower + 3).to_integer
				create token_probability.make (token, probability, bad_count, good_count)
				tokens.put (token_probability, token)
				file.read_line
				line := line + 1
			end
			file.close
		end

	save (a_file_name: STRING; a_minimum_occurrences: INTEGER) is
			-- Save probability table to a file.
		require
			a_file_name_not_empty: a_file_name /= Void and then not a_file_name.is_empty
			minimum_occurrences_positive: a_minimum_occurrences >= 1
		local
			file: STDC_TEXT_FILE
			data_point: TOKEN_PROBABILITY
			old_file_name: STRING
			new_file_name: STRING
		do
			old_file_name := a_file_name + ".old"
			new_file_name := a_file_name + ".new"
			create file.create_write (new_file_name)
			-- First line is spam_messages_count and nonspam_messages_count
			file.put_string (spam_messages_count.out)
			file.write_character (' ')
			file.put_string (nonspam_messages_count.out)
			file.write_character ('%N')
			-- Second line is x, the number to be used when there are
			-- no data points.
			file.put_string (no_data_points_value.out)
			file.write_character ('%N')
			-- Third line is number of tokens
			file.put_string (tokens.count.out)
			file.write_character ('%N')
			-- Rest of lines are tokens
			from
				tokens.start
			until
				tokens.after
			loop
				data_point := tokens.item_for_iteration
				if data_point.bad_count + data_point.good_count >= a_minimum_occurrences then
					file.put_string (data_point.probability.out)
					file.write_character (' ')
					file.put_string (data_point.token)
					file.write_character (' ')
					file.put_string (data_point.bad_count.out)
					file.write_character (' ')
					file.put_string (data_point.good_count.out)
					file.write_character ('%N')
				end
				tokens.forth
			end
			file.close
			if is_existing (a_file_name) then
				if is_existing (old_file_name) then
					remove_file (old_file_name)
				end
				rename_to (a_file_name, old_file_name)
			end
			rename_to (new_file_name, a_file_name)
		end


feature -- Access

	nonspam_messages_count: INTEGER
			-- Number of messages scanned for non-spam tokens.

	spam_messages_count: INTEGER
			-- Number of messages scanned for spam tokens.

	tokens: DS_HASH_TABLE [TOKEN_PROBABILITY, STRING]
			-- Token with probabilities.

	no_data_points_value: DOUBLE
			-- The value to use when a token in a message does not occur
			-- in `tokens'.
			-- The x in Gary Robinson's formula'.s


feature {NONE} -- Error handling

	assert (line: INTEGER; assert_msg: STRING; success: BOOLEAN) is
			-- If not `success', dump error info to stderr and exit.
		require
			positive_line: line > 0
			assert_msg_not_empty: assert_msg /= Void and then not assert_msg.is_empty
		do
			if not success then
				stderr.put_string ("Format violation: ")
				stderr.put_string (assert_msg)
				stderr.put_string (" at line ")
				stderr.put_string (line.out)
				stderr.put_string (".%N")
				exit_with_failure
			end
		ensure
			exit_on_failure: success
		end


invariant

	tokens_not_void: tokens /= Void
	no_data_points_value_is_probability: no_data_points_value >= 0 and no_data_points_value <= 1
	nonspam_messages_count_positive: nonspam_messages_count > 0
	spam_messages_count_positive: spam_messages_count > 0

end
