indexing

	description: "EMC configuration contents."

	author: "Berend de Boer"


class

	CONFIGURATION


creation

	make


feature {NONE} -- Initialization

	make (a_stream: KI_CHARACTER_INPUT_STREAM) is
			-- Load configuration from `a_stream'.
			-- If the configuration cannot be read correctly,
			-- `is_correct' will be False, and `last_error' contains the
			-- error message
		require
			a_stream_not_void: a_stream /= Void
			is_open_read: a_stream.is_open_read
		local
			configuration_parser: CONFIGURATION_PARSER
		do
			minimum_occurrences := 1
			create {DS_LINKED_LIST [MAILBOX]} spam_mailboxes.make
			create {DS_LINKED_LIST [MAILBOX]} nonspam_mailboxes.make
			set_strength (1.0)
			set_password_from_environment_variable
			set_scan_mailbox (default_scan_mailbox)
			create configuration_parser.make
			configuration_parser.parse_from_stream (a_stream, Current)
			is_correct := configuration_parser.is_correct
			if is_correct then
				last_error := Void
			else
				last_error := configuration_parser.last_error
			end
		end


feature -- Configuration reading

	is_correct: BOOLEAN
			-- Was reading the configuration from file successful?

	last_error: STRING
			-- Any error message if not `is_correct'.


feature -- Mail server

	host: STRING
			-- Name of server to access.

	password: STRING
			-- Password.

	port: INTEGER
			-- Port; zero if not set. Will use default port in that case.

	login_name: STRING
			-- Login name.

	secure: BOOLEAN
			-- Must connection to the server be secure?


	set_host (a_host: STRING) is
		require
			a_host_not_empty: a_host /= Void and then not a_host.is_empty
		do
			host := a_host
		end

	set_login_name (a_login_name: STRING) is
		require
			login_name_not_empty: a_login_name /= Void and then not a_login_name.is_empty
		do
			login_name := a_login_name
		end

	set_password (a_password: STRING) is
		require
			password_not_empty: a_password /= Void and then not a_password.is_empty
		do
			password := a_password
		end

	set_password_from_environment_variable is
			-- Set `password' from the environment variable
			-- IMAP4_PASSWORD if this is defined.
		local
			password_env: STDC_ENV_VAR
		do
			create password_env.make ("IMAP4_PASSWORD")
			if not password_env.value.is_empty then
				password := password_env.value
			end
		end

	set_port (a_port: INTEGER) is
		require
			valid_port: a_port >= 1 and then a_port <= 65535
		do
			port := a_port
		ensure
			port_set: port = a_port
		end

	set_secure (a_secure: BOOLEAN) is
		do
			secure := a_secure
		ensure
			secure_set: secure = a_secure
		end


feature -- Options used in learning mode

	minimum_occurrences: INTEGER
			-- Minimum number of occurrences for a token to appear in the
			-- token table.
			-- The default value is 1.

	spam_mailboxes: DS_LIST [MAILBOX]

	nonspam_mailboxes: DS_LIST [MAILBOX]

	set_minimum_occurrences (a_minimum_occurrences: INTEGER) is
			-- Set `minimum_occurrences'.
		require
			minimum_occurrences_positive: a_minimum_occurrences >= 1
		do
			minimum_occurrences := a_minimum_occurrences
		ensure
			minimum_occurrences_set: minimum_occurrences = a_minimum_occurrences
		end


feature -- Status

	has_nonspam_mailbox (a_mailbox_name: STRING): BOOLEAN is
			-- Does `nonspam_mailboxes' have the mail box `a_mailbox_name'?
		require
			a_mailbox_name_not_empty: a_mailbox_name /= Void and then not a_mailbox_name.is_empty
		do
			from
				nonspam_mailboxes.start
			until
				Result or else
				nonspam_mailboxes.after
			loop
				Result := nonspam_mailboxes.item_for_iteration.name.is_equal (a_mailbox_name)
				nonspam_mailboxes.forth
			end
		end

	has_spam_mailbox (a_mailbox_name: STRING): BOOLEAN is
			-- Does `spam_mailboxes' have the mail box `a_mailbox_name'?
		require
			a_mailbox_name_not_empty: a_mailbox_name /= Void and then not a_mailbox_name.is_empty
		do
			from
				spam_mailboxes.start
			until
				Result or else
				spam_mailboxes.after
			loop
				Result := spam_mailboxes.item_for_iteration.name.is_equal (a_mailbox_name)
				spam_mailboxes.forth
			end
		end


feature -- Options used in classification

	scan_mailbox: MAILBOX
			-- Email inbox to scan for spam and nonspam messages

	spam_mailbox: STRING
			-- Mailbox to move spam to;
			-- If Void, spam is not moved.

	nonspam_mailbox: STRING
			-- Mailbox to move nonspam to;
			-- If Void, nonspam is not moved.

	strength: DOUBLE
			-- The magic `s' in Gary Robinson's formula;
			-- It is the strength of your believe that the probability
			-- assigned to new words to rate their spaminess is correct.

	threshold: DOUBLE
			-- The value above which we consider something to be spam

	lower_minimum_deviation,
	upper_minimum_deviation: DOUBLE
			-- The range around `threshold' in which to discard words;
			-- those words probably do not contribute significantly to
			-- rating spaminess of the email or might lessen the
			-- spaminess rate of it.

	new_word_spaminess: DOUBLE
			-- The spaminess to assign to words not in the token list

	set_scan_mailbox (a_mailbox: MAILBOX) is
		require
			mailbox_not_void: a_mailbox /= Void
		do
			scan_mailbox := a_mailbox
		end

	set_spam_mailbox (a_mailbox: STRING) is
			-- Set `spam_mailbox' to Void if `a_mailbox' is Void or
			-- empty, else set it to `a_mailbox'.
		do
			if a_mailbox = Void or else a_mailbox.is_empty then
				spam_mailbox := Void
			else
				spam_mailbox := a_mailbox
			end
		end

	set_nonspam_mailbox (a_mailbox: STRING) is
			-- Set `nonspam_mailbox' to Void if `a_mailbox' is Void or
			-- empty, else set it to `a_mailbox'.
		do
			if a_mailbox = Void or else a_mailbox.is_empty then
				nonspam_mailbox := Void
			else
				nonspam_mailbox := a_mailbox
			end
		end

	set_strength (a_strength: DOUBLE) is
		require
			strength_valid: a_strength >= -1.0 and a_strength <= 1.0
		do
			strength := a_strength
		ensure
			strength_set: strength = a_strength
		end

	set_threshold (a_threshold: DOUBLE) is
		require
			threshold_valid: a_threshold >= -1.0 and a_threshold <= 1.0
		do
			threshold := a_threshold
		ensure
			threshold_set: threshold = a_threshold
		end

	set_lower_minimum_deviation (a_lower: DOUBLE) is
		require
			lower_valid: a_lower >= -1 and a_lower <= 1.0
		do
			lower_minimum_deviation := a_lower
			if lower_minimum_deviation > upper_minimum_deviation then
				upper_minimum_deviation := lower_minimum_deviation
			end
		ensure
			lower_minimum_deviation_set: lower_minimum_deviation = a_lower
		end

	set_upper_minimum_deviation (an_upper: DOUBLE) is
		require
			upper_valid: an_upper >= -1 and an_upper <= 1.0
		do
			upper_minimum_deviation := an_upper
			if lower_minimum_deviation > upper_minimum_deviation then
				lower_minimum_deviation := upper_minimum_deviation
			end
		ensure
			upper_minimum_deviation_set: upper_minimum_deviation = an_upper
		end

	set_new_word_spaminess (a_new_word_spaminess: DOUBLE) is
			-- Set the spaminess to assign to words not in the token list.
		require
			valid_probability: a_new_word_spaminess >= -1 and a_new_word_spaminess <= 1.0
		do
			new_word_spaminess := a_new_word_spaminess
		ensure
			new_word_spaminess_set: new_word_spaminess = a_new_word_spaminess
		end


feature {NONE} -- Implementation

	default_scan_mailbox: MAILBOX is
		once
			create Result.make (
				"INBOX",
				0,
				0,
				0,
				1,
				0)
		ensure
			default_scan_mailbox_not_void: default_scan_mailbox /= Void
		end


invariant

	minimum_occurrences_positive: minimum_occurrences > 0
	spam_mailboxes_not_void: spam_mailboxes /= Void
	nonspam_mailboxes_not_void: nonspam_mailboxes /= Void

	password_void_or_not_empty: password = Void or else not password.is_empty
	scan_mailbox_not_void: scan_mailbox /= void

	spam_mailbox_void_or_not_empty: spam_mailbox = Void or else not spam_mailbox.is_empty
	nonspam_mailbox_void_or_not_empty: nonspam_mailbox = Void or else not nonspam_mailbox.is_empty
	valid_strength: strength >= 0 and strength <= 1
	valid_threshold: threshold >= -1.0 and threshold <= 1.0
	valid_lower_minimum_deviation: lower_minimum_deviation >= 0 and lower_minimum_deviation <= 1.0
	valid_upper_minimum_deviation: upper_minimum_deviation >= 0 and upper_minimum_deviation <= 1.0
	valid_minimum_deviation_range: lower_minimum_deviation <= upper_minimum_deviation

end
