indexing

	description: "Various routines to scan mail into tokens."


class

	TOKENIZE_ROUTINES


inherit

	ANY

	EPX_CURRENT_PROCESS
		export
			{NONE} all
		end

	EPX_MIME_TYPE_NAMES
		export
			{NONE} all
		end

	EPX_MIME_FIELD_NAMES
		export
			{NONE} all
		end

	EPX_MIME_PARAMETER_NAMES
		export
			{NONE} all
		end


feature -- Body tokenizing

	tokenize_text_bodies (a_tokenizer: TOKENIZER; a_message_number: INTEGER; a_part: EPX_MIME_PART) is
			-- Recursively parse all bodies in `a_part' if it is a text body.
		require
			tokenizer_not_void: a_tokenizer /= Void
			part_not_void: a_part /= Void
		local
			i: INTEGER
			s: STRING
		do
			if a_part.body.is_multipart then
				from
					i := 1
				until
					i > a_part.body.parts_count
				loop
					tokenize_text_bodies (a_tokenizer, a_message_number, a_part.body.part (i))
					i := i + 1
				end
			else
				if is_text_body (a_part.header.content_type) then
					s := a_part.body.as_plain_text
					if not s.is_empty then
						-- Perhaps here to remove SpamAssassin text?
						-- See for example msg: 5587
						a_tokenizer.tokenize (s)
					end
				else
					tokenize_characteristics (a_tokenizer, a_part.header)
					show_skipped_body (a_message_number, a_part)
				end
			end
		end


feature -- Header tokenizing

	sanitize_subject (a_subject: STRING): STRING is
			-- Clean up subject some what. For example remove the *SPAM*
			-- prefix added by spamassassin, and the "Re:" prefix.
		require
			a_subject_not_void: a_subject /= Void
		local
			p: INTEGER
			starts_with_blanks,
			starts_with_re: BOOLEAN
		do
			p := a_subject.substring_index (once_spam, 1)
			if p /= 0 then
				p := p + once_spam.count
				from
				until
					p > a_subject.count or else
					a_subject.item (p) = ' '
				loop
					p := p + 1
				end
				Result := a_subject.substring (p+1, a_subject.count)
			else
				Result := a_subject
			end

			-- removing blanks
			starts_with_blanks :=
				Result.count > 0 and then
				Result.item (1) = ' '
			if starts_with_blanks then
				if Result = a_subject then
					Result := Result.twin
				end
				Result.left_adjust
				Result.right_adjust
			end

			-- Remove starting "Re:"
			starts_with_re :=
				Result.count >= 3 and then
				(Result.item (1) = 'R' or else Result.item (1) = 'r') and then
				(Result.item (2) = 'E' or else Result.item (2) = 'e') and then
				(Result.item (3) = ':')
			if starts_with_re then
				if Result = a_subject then
					Result := Result.twin
				end
				Result.remove_head (3)
				Result.left_adjust
			end
		ensure
			sanitized_subject_not_void: Result /= Void
		end

	tokenize_header (a_tokenizer: TOKENIZER; a_header: EPX_MIME_HEADER) is
			-- Tokenize interesting parts of `a_header' as its subject.
			-- Header usually contains a lot of repetitive, one time
			-- information like message id's or host names that are not
			-- very discriminating.
			-- I think it's best if only user visible information is
			-- added, i.e. that information that a human would use to
			-- make the decision if it's spam or not.
			-- Other fields should not be mixed with human readable
			-- tokens, but added as special tokens, for example
			-- $MailClient or so.
		require
			tokenizer_not_void: a_tokenizer /= Void
			header_not_void: a_header /= Void
		local
			subject: STRING
		do
			-- Append subject to text of bodies.
			if a_header.has (field_name_subject) then
				subject := sanitize_subject (a_header.item (field_name_subject).value)
				if not subject.is_empty then
					a_tokenizer.tokenize (subject)
				end
			end
			--tokenize_field (a_tokenizer, a_header, field_name_x_mail)

			tokenize_characteristics (a_tokenizer, a_header)
		end


feature {NONE} -- Special tokens

	tokenize_bad_body (a_tokenizer: TOKENIZER) is
		require
			tokenizer_not_void: a_tokenizer /= Void
		do
			a_tokenizer.tokenize_special (once_bad_body)
		end

	tokenize_bad_header (a_tokenizer: TOKENIZER) is
		require
			tokenizer_not_void: a_tokenizer /= Void
		do
			a_tokenizer.tokenize_special (once_bad_header)
		end

	tokenize_characteristics (a_tokenizer: TOKENIZER; a_header: EPX_MIME_HEADER) is
			-- Extract certain characteristics from a mail header that
			-- might indicate spaminess.
		require
			tokenizer_not_void: a_tokenizer /= Void
			header_not_void: a_header /= Void
		do
			tokenize_charset (a_tokenizer, a_header)
			tokenize_content_disposition (a_tokenizer, a_header)
		end

	tokenize_charset (a_tokenizer: TOKENIZER; a_header: EPX_MIME_HEADER) is
			-- Certain character sets are probably never used by
			-- someone. Add them as special tokens to the token list.
		require
			tokenizer_not_void: a_tokenizer /= Void
			header_not_void: a_header /= Void
		local
			charset: STRING
			charset_correct: BOOLEAN
		do
			if a_header.content_type /= Void then
				a_header.content_type.parameters.search (parameter_name_charset)
				if a_header.content_type.parameters.found then
					charset_correct :=
						not a_header.content_type.parameters.found_item.value.is_empty and then
						not a_header.content_type.parameters.found_item.value.has (' ') and then
						not a_header.content_type.parameters.found_item.value.has ('%T')
					if charset_correct then
						charset := a_header.content_type.parameters.found_item.value
						a_tokenizer.tokenize_special (charset.as_lower)
					end
				end
			end
		end

	tokenize_content_disposition (a_tokenizer: TOKENIZER; a_header: EPX_MIME_HEADER) is
			-- Extract filename from Content-Disposition if any.
		require
			tokenizer_not_void: a_tokenizer /= Void
			header_not_void: a_header /= Void
		local
			filename: STDC_PATH
			s: STRING
			i: INTEGER
		do
			if a_header.content_disposition /= Void then
				a_header.content_disposition.parameters.search (parameter_name_filename)
				if a_header.content_disposition.parameters.found then
					create filename.make_from_string (a_header.content_disposition.parameters.found_item.value)
					filename.parse (executable_suffices)
					-- Parse file name, but distinguish parts separated by '_'.
					s := filename.basename
					from
						i := 1
					variant
						s.count - (i - 1)
					until
						i > s.count
					loop
						inspect s.item (i)
						when '_', '-' then
							s.put (' ', i)
						else
							-- ok
						end
						i := i + 1
					end
					if not s.is_empty then
						a_tokenizer.tokenize (s)
					end
					if not filename.suffix.is_empty then
						a_tokenizer.tokenize_special (once_suffix + filename.suffix.as_lower)
					end
				end
			end
		end


feature {NONE} -- Implementation

	executable_suffices: ARRAY [STRING] is
			-- Well-known executable suffices on Windows platform
		once
			Result := <<".cmd", ".com", ".exe", ".scr", ".bat", ".pif">>
		end

	is_text_body (a_content_type: EPX_MIME_FIELD_CONTENT_TYPE): BOOLEAN is
			-- Is this a Content-Type that indicates some form of a text
			-- body?
		local
			type,
			subtype: STRING
		do
			Result := a_content_type = Void
			if not Result then
				type := a_content_type.type.as_lower
				subtype := a_content_type.subtype.as_lower
				Result :=
					(type.is_equal (mime_type_text)) or else
					(type.is_equal (mime_type_message) and then
					 (subtype.is_equal (mime_subtype_rfc822) or else subtype.is_equal (mime_subtype_delivery_status))) or else
					(type.is_equal (mime_type_application) and then
					 subtype.is_equal (mime_subtype_x_tex))

			end
		end

	tokenize_field (
			a_tokenizer: TOKENIZER;
			a_header: EPX_MIME_HEADER;
			a_field_name: STRING) is
		require
			tokenizer_not_void: a_tokenizer /= Void
			header_not_void: a_header /= Void
			a_field_name_not_empty: a_field_name /= Void and then not a_field_name.is_empty
		local
			contents: STRING
		do
			if a_header.has (a_field_name) then
				contents := a_header.item (a_field_name).value
				if not contents.is_empty then
					a_tokenizer.tokenize (contents)
				end
			end
		end


feature {NONE} -- Logging

	ignorable_content_types: DS_HASH_TABLE [STRING, STRING] is
			-- Known Content-Types that contain no text.
			-- Used for debugging to be warned about possible
			-- Content-Types that can be parsed as text.
		once
			create Result.make (32)
			Result.put_last (Void, "audio/midi")
			Result.put_last (Void, "audio/x-midi")
			Result.put_last (Void, "audio/x-wav")
			Result.put_last (Void, "image/bmp")
			Result.put_last (Void, "image/gif")
			Result.put_last (Void, "image/jpeg")
			Result.put_last (Void, "image/pjpeg")
			Result.put_last (Void, "image/tiff")
			Result.put_last (Void, "application/hta")
			Result.put_last (Void, "application/ms-tnef")
			Result.put_last (Void, "application/msword")
			Result.put_last (Void, "application/octet-stream")
			Result.put_last (Void, "application/pdf")
			Result.put_last (Void, "application/pgp-signature")
			Result.put_last (Void, "application/postscript")
			Result.put_last (Void, "application/pkcs7-signature")
			Result.put_last (Void, "application/vnd.ms-excel")
			Result.put_last (Void, "application/x-compressed")
			Result.put_last (Void, "application/x-emacs-lisp")
			Result.put_last (Void, "application/x-gtar")
			Result.put_last (Void, "application/x-gzip")
			Result.put_last (Void, "application/x-java-applet")
			Result.put_last (Void, "application/x-msdownload")
			Result.put_last (Void, "application/x-tgz")
			Result.put_last (Void, "application/x-zip-compressed")
			Result.put_last (Void, "application/zip")
		ensure
			ignorable_content_types_not_void: ignorable_content_types /= Void
		end

	show_skipped_body (a_message_number: INTEGER; a_part: EPX_MIME_PART) is
		-- Some info to stderr when a body is skipped. Info is only
		-- written when we don't know if it can be ignored.
		require
			part_not_void: a_part /= Void
		local
			spam_assassin_rewrote_it: BOOLEAN
			s: STRING
		do
			if not ignorable_content_types.has (a_part.header.content_type.value.as_lower) then
				spam_assassin_rewrote_it :=
					a_part.header.has (field_name_x_spam_checker_version) and then
					a_part.header.item (field_name_x_spam_checker_version).value.has_substring ("SpamAssassin")
				if not spam_assassin_rewrote_it then
					stderr.write_string ("Skipping body part of message ")
					stderr.write_string (a_message_number.out)
					stderr.write_string (" with Content-Type ")
					stderr.write_string (a_part.header.content_type.value)
					stderr.write_string ("%N")
					debug ("emc")
						stderr.put_string ("==========begin==========%N")
						if a_part.header.has (field_name_subject) then
							stderr.write_string (a_part.header.item (field_name_subject).value)
							stderr.write_string ("%N")
						end
						stderr.write_string ("%N")
						s := a_part.body.as_string
						if s.count <= 1024 then
							stderr.write_string (s)
						else
							stderr.write_string (s.substring (1, 1024))
							stderr.write_string ("...%N")
						end
						stderr.put_string ("==========end==========%N%N")
					end
				end
			end
		end


feature {NONE} -- Field names

	field_name_x_mail: STRING is "X-Mailer"

	field_name_x_spam_checker_version: STRING is "X-Spam-Checker-Version"


feature {NONE} -- Special tokens

	once_bad_body: STRING is "bad-body"
	once_bad_header: STRING is "bad-header"

	once_suffix: STRING is "suffix:"


feature {NONE} -- Once strings

	once_spam: STRING is "*SPAM*"


end
